from .log import StructuredDataLogger
from .application import Application
from .abc.service import Service
from .abc.module import Module
from .abc.singleton import Singleton
from .config import Config
from .pubsub import subscribe, PubSub

from .socket import StreamSocketServerService



import logging
import asyncio
import argparse
import itertools
import os
import signal
import platform

from .config import Config
from .abc.singleton import Singleton
from .log import _setup_logging, _loop_exception_handler

from .metrics import Metrics

# Importing the Win API library
if platform.system() == "Windows":
	try:
		import win32api
	except ModuleNotFoundError:
		win32api = None
else:
	win32api = None

#

L = logging.getLogger(__name__)

#

class Application(metaclass=Singleton):


	Description = ""


	def __init__(self):

		# Parse command line
		self.parse_args()

		# Load configuration
		Config._load()

		# Setup logging
		_setup_logging()

		# Configure event loop
		self.Loop = asyncio.get_event_loop()
		self.Loop.set_exception_handler(_loop_exception_handler)
		if Config["general"]["verbose"] == "True":
			self.Loop.set_debug(True)

		
		try:
			# Signals are not available on Windows
			self.Loop.add_signal_handler(signal.SIGINT, self.stop)
		except NotImplementedError:
			# Checking if the program runs on Windows
			if any(platform.win32_ver()):
				if win32api is not None:
					callBack = self.stop
					# Adding a handler to listen to the interrupt event
					def handler(type):
						callBack()
						return True
					win32api.SetConsoleCtrlHandler(handler, True)

		try:
			self.Loop.add_signal_handler(signal.SIGTERM, self.stop)
		except NotImplementedError:
			pass

		self._stop_event = asyncio.Event(loop = self.Loop)
		self._stop_event.clear()
		self._stop_counter = 0

		from .pubsub import PubSub
		self.PubSub = PubSub(self)
		self.Metrics = Metrics(self)

		self.Modules = []
		self.Services = {}

		# Comence init-time governor
		L.info("Initializing ...")
		finished_tasks, pending_tasks = self.Loop.run_until_complete(asyncio.wait(
			[
				self.initialize(),
				self._init_time_governor(asyncio.Future()),
			],
			return_when = asyncio.FIRST_EXCEPTION
		))
		for task in finished_tasks:
			# This one also raises exceptions from futures, which is perfectly ok
			task.result()
		if len(pending_tasks) > 0:
			raise RuntimeError("Failed to fully initialize. Here are pending tasks: {}".format(pending_tasks))


	def parse_args(self):
		'''
		This method can be overriden to adjust argparse configuration 
		'''

		parser = argparse.ArgumentParser(
			formatter_class=argparse.RawDescriptionHelpFormatter,
			description=self.Description,
		)
		parser.add_argument('-c', '--config', help='Path to configuration file (default: %(default)s)', default=Config._default_values['general']['config_file'])
		parser.add_argument('-v', '--verbose', action='store_true', help='Print more information (enable debug output)')

		args = parser.parse_args()
		if args.config is not None:
			Config._default_values['general']['config_file'] = args.config

		if args.verbose:
			Config._default_values['general']['verbose'] = True


	def run(self):
		# Comence run-time and application main() function
		L.info("Running ...")
		self._stop_event.clear()
		finished_tasks, pending_tasks = self.Loop.run_until_complete(asyncio.wait(
			[
				self.main(),
				self._run_time_governor(asyncio.Future()),
			],
			return_when = asyncio.FIRST_EXCEPTION
		))
		for task in finished_tasks:
			try:
				task.result()
			except Exception:
				L.exception("Exception in {}".format(task))

		#TODO: Process pending_tasks tasks from above

		# Comence exit-time
		L.info("Exiting ...")
		finished_tasks, pending_tasks = self.Loop.run_until_complete(asyncio.wait(
			[
				self.finalize(),
				self._exit_time_governor(asyncio.Future()),
			],
			return_when = asyncio.FIRST_EXCEPTION
		))
		for task in finished_tasks:
			try:
				task.result()
			except Exception:
				L.exception("Exception in {}".format(task))

		#TODO: Process pending_tasks tasks from above


		self.Loop.close()

		try:
			# EX_OK code is not available on Windows
			return os.EX_OK
		except AttributeError:
			return 0


	def stop(self):
		self._stop_event.set()
		self._stop_counter += 1
		if self._stop_counter >= 3:
			L.fatal("Emergency exit")
			try:
				# EX_SOFTWARE code is not available on Windows
				return os._exit(os.EX_SOFTWARE)
			except AttributeError:
				return os._exit(0)


	# Modules

	def add_module(self, module_class):
		""" Load a new module. """

		module = module_class(self)
		self.Modules.append(module)
	
		asyncio.ensure_future(module.initialize(self), loop=self.Loop)

	# Services

	def get_service(self, service_name):
		""" Get a new service by its name. """

		try:
			return self.Services[service_name]
		except KeyError:
			pass

		L.error("Cannot find service '{}' - not registered?".format(service_name))
		raise KeyError("Cannot find service '{}'".format(service_name))


	def register_service(self, service_name, service):
		""" Register a new service using its name. """

		if service_name in self.Services:
			L.error("Service '{}' already registered (existing:{} new:{})"
					.format(service_name, self.Services[service_name], service))
			raise RuntimeError("Service {} already registered".format(service_name))

		self.Services[service_name] = service

		asyncio.ensure_future(service.initialize(self), loop=self.Loop)


	# Lifecycle callback

	async def initialize(self):
		pass

	async def main(self):
		pass

	async def finalize(self):
		pass

	# Governors

	async def _init_time_governor(self, future):
		self.PubSub.publish("Application.init!")
		future.set_result("initialize")


	async def _run_time_governor(self, future):
		timeout = Config.getint('general', 'tick_period')
		try:
			self.PubSub.publish("Application.run!")

			# Wait for stop event & tick in meanwhile
			for cycle_no in itertools.count(1):
				try:
					await asyncio.wait_for(self._stop_event.wait(), timeout=timeout)
					break
				except asyncio.TimeoutError:
					self.Metrics.add("Application.tick", 1)
					self.PubSub.publish("Application.tick!")
					if (cycle_no % 10) == 0: self.PubSub.publish("Application.tick/10!")
					if (cycle_no % 60) == 0: self.PubSub.publish("Application.tick/60!")
					if (cycle_no % 300) == 0: self.PubSub.publish("Application.tick/300!")
					if (cycle_no % 600) == 0: self.PubSub.publish("Application.tick/600!")
					if (cycle_no % 1800) == 0: self.PubSub.publish("Application.tick/1800!")
					if (cycle_no % 3600) == 0: self.PubSub.publish("Application.tick/3600!")
					if (cycle_no % 43200) == 0: self.PubSub.publish("Application.tick/43200!")
					if (cycle_no % 86400) == 0: self.PubSub.publish("Application.tick/86400!")
					continue

		finally:
			future.set_result("run")


	async def _exit_time_governor(self, future):
		self.PubSub.publish("Application.exit!")

		# Finalize services
		futures = []
		for service in self.Services.values():
			nf = asyncio.ensure_future(service.finalize(self), loop=self.Loop)
			futures.append(nf)
		if len(futures) > 0:
			await asyncio.wait(futures, return_when=asyncio.ALL_COMPLETED)
			# TODO: Handle expections (if needed) - probably only print them

		# Finalize modules
		futures = []
		for module in self.Modules:
			nf = asyncio.ensure_future(module.finalize(self), loop=self.Loop)
			futures.append(nf)
		if len(futures) > 0:
			await asyncio.wait(futures, return_when=asyncio.ALL_COMPLETED)
			# TODO: Handle expections (if needed) - probably only print them

		future.set_result("exit")



import os
import sys
import glob
import logging
import configparser

#

L = logging.getLogger(__name__)

#

class ConfigParser(configparser.ConfigParser):

	_default_values = {

		'general': {
			'verbose': os.environ.get('ASAB_VERBOSE', False),
			'config_file': os.environ.get('ASAB_CONFIG', ''),
			'tick_period': 1, # In seconds
		},

		"logging:rfc5424": {
			"sd_id": "sd",
		},

		"logging:console": {
			"format": "%%(asctime)s %%(levelname)s %%(name)s %%(struct_data)s: %%(message)s",
			"datefmt": "%%d-%%b-%%Y %%H:%%M:%%S.%%f",
		},

	}


	def add_defaults(self, dictionary):
		""" Add defaults to a current configuration """

		for section, keys in dictionary.items():
			section = str(section)

			if section not in self._sections:
				try:
					self.add_section(section)
				except ValueError:
					if self._strict:
						raise

			for key, value in keys.items():

				key = self.optionxform(str(key))
				if key in self._sections[section]:
					# Value exists, no default needed
					continue 

				if value is not None:
					value = str(value)

				if "$" in value:
					self.set(section, key, os.path.expandvars(value))				
				else:
					self.set(section, key, value)


	def _load(self):
		""" This method should be called only once, any subsequent call will lead to undefined behaviour """

		config_fname = ConfigParser._default_values['general']['config_file']
		if config_fname != '':
			if not os.path.isfile(config_fname):
				print("Config file '{}' not found".format(config_fname), file=sys.stderr)
				sys.exit(1)

			self.read(config_fname)

		includes = self.get('general', 'include', fallback='')
		for include_glob in includes.split(os.pathsep):
			for include in glob.glob(include_glob):
				self.read(include)

		self.add_defaults(ConfigParser._default_values)

		# Deals with environment variables
		for each_section in self.sections():
			for (each_key, each_val) in self.items(each_section):
				if "$" in each_val:
					self.set(each_section, each_key, os.path.expandvars(each_val))

###

Config = ConfigParser()



import logging
import asyncio
import inspect
import functools

#

L = logging.getLogger(__name__)

#


class PubSub(object):

	def __init__(self, app):
		self.subscribers = {}
		self.Loop = app.Loop


	def subscribe(self, message_type, callback):
		"""
		Subscribe a subscriber to the an message type.
		It could be even plain function, method or its coroutine variant (then it will be delivered in a dedicated future)
		"""

		def _deliver_async(loop, callback, message_type, *args, **kwargs):
			asyncio.ensure_future(callback(message_type, *args, **kwargs), loop=loop)

		# If subscribe is a coroutine (async def), then wrap delivery in 
		if inspect.iscoroutinefunction(callback):
			callback = functools.partial(_deliver_async, self.Loop, callback)

		if message_type not in self.subscribers:
			self.subscribers[message_type] = set([callback])
		else:
			self.subscribers[message_type].add(callback)


	def subscribe_all(self, obj):
		"""
		Find all @asab.subscribe decorated methods on the obj and do subscription
		"""
		for member_name in dir(obj):
			member = getattr(obj, member_name)
			message_types = getattr(member, 'asab_pubsub_subscribe_to_message_types', None)
			if message_types is not None:
				for message_type in message_types:
					self.subscribe(message_type, member)


	def unsubscribe(self, message_type, callback):
		""" Remove a subscriber of an message type from the set. """

		callback_set = self.subscribers.get(message_type)
		if callback_set is None:
			L.warning("Message type subscription '{}'' not found.".format(message_type))
			return
		else:
			try:
				callback_set.remove(callback)
			except KeyError:
				L.warning("Subscriber '{}'' not found for the message type '{}'.".format(message_type, callback))


	def publish(self, message_type, *args, **kwargs):
		""" Notify subscribers of an message type. Including arguments. """

		callback_set = self.subscribers.get(message_type)
		if callback_set is None:
			return

		asynchronously = kwargs.pop('asynchronously', False)

		if asynchronously:
			for callback in callback_set:
				self.Loop.call_soon(functools.partial(callback, message_type, *args, **kwargs))

		else:
			for callback in callback_set:
				callback(message_type, *args, **kwargs)

###

class subscribe(object):

	'''
	Decorator

	Usage:
	
	@asab.subscribe("tick")
	def on_tick(self, message_type):
	print("Service tick")
	'''

	def __init__(self, message_type):
		self.message_type = message_type

	def __call__(self, f):
		if getattr(f, 'asab_pubsub_subscribe_to_message_types', None) is None:
			f.asab_pubsub_subscribe_to_message_types = [self.message_type]
		else:
			f.asab_pubsub_subscribe_to_message_types.append(self.message_type)
		return f





class Singleton(type):

	"""
	The singleton pattern is a software design pattern that restricts the instantiation of a class to one object.
	More at https://en.wikipedia.org/wiki/Singleton_pattern
	Implementation idea is from https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

	Usage:

	class MyClass(metaclass=Singleton):
		...

	"""

	_instances = {}

	def __call__(cls, *args, **kwargs):
		if cls not in cls._instances:
			cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
		return cls._instances[cls]



import abc

class Service(abc.ABC):
	"""
	Abstract service class
	"""

	def __init__(self, app):
		pass


	# Lifecycle

	async def initialize(self, app):
		pass

	async def finalize(self, app):
		pass



import logging
import collections

#

L = logging.getLogger(__name__)

#


class Metrics(collections.MutableMapping):
	"""
	Manage scalar metrics of the application
	"""

	def __init__(self, app):
		self.metrics = {}

	def add(self, metric_name, value=1):
		""" Add a value to the specified metric. """

		if metric_name not in self.metrics:
			self.metrics[metric_name] = 0
		self.metrics[metric_name] += value

	def set(self, metric_name, value=0):
		""" Set a value of a specified metric. """

		self.metrics[metric_name] = value

	def keys(self):
		""" Get metric names as keys. """

		return self.metrics.keys()

	def __getitem__(self, key):
		if key not in self.metrics:
			return None
		return self.metrics.__getitem__(key)

	def __setitem__(self, key, value):
		return self.metrics.__setitem__(key, value)

	def __delitem__(self, key):
		if key not in self.metrics:
			return
		return self.metrics.__delitem__(key)

	def __iter__(self):
		return self.metrics.__iter__()

	def __len__(self):
		return self.metrics.__len__()



from setuptools import setup
from setuptools import find_packages

setup(
	name='asab',
	version='18.03-beta1',
	description='Asynchronous Server Application Boilerplate',
	url='https://github.com/TeskaLabs/asab',
	author='TeskaLabs Ltd',
	author_email='support@teskalabs.com',
	classifiers=[
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.6',
	],
	keywords='asynchornous server application boilerplate',
	packages=find_packages(exclude=['module_sample']),
	project_urls={
		'Source': 'https://github.com/TeskaLabs/asab'
	},
)



import os
import sys
import logging
import logging.handlers
import traceback
import time
import datetime
import pprint
import socket
from .config import Config


def _setup_logging():

	root_logger = logging.getLogger()
	if not root_logger.hasHandlers():
		
		# Add console logger
		#TODO: Don't initialize this when not on console
		h = logging.StreamHandler(stream=sys.stderr)
		h.setFormatter(StructuredDataFormatter(
			fmt = Config["logging:console"]["format"],
			datefmt = Config["logging:console"]["datefmt"],
			sd_id = Config["logging:rfc5424"]["sd_id"],
		))
		h.setLevel(logging.DEBUG)
		root_logger.addHandler(h)

		#TODO: If configured, initialize syslog

	else:
		root_logger.warning("Logging seems to be already configured. Proceed with caution.")

	if Config["general"]["verbose"] == "True":
		root_logger.setLevel(logging.DEBUG)
	else:
		root_logger.setLevel(logging.WARNING)


class StructuredDataLogger(logging.Logger):

	def _log(self, level, msg, args, exc_info=None, struct_data=None, extra=None, stack_info=False):
		if struct_data is not None:
			if extra is None: extra = dict()
			extra['_struct_data'] = struct_data

		super()._log(level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info)

logging.setLoggerClass(StructuredDataLogger)


class StructuredDataFormatter(logging.Formatter):

	def __init__(self, fmt=None, datefmt=None, style='%', sd_id=None):
		super().__init__(fmt, datefmt, style)
		self.sd_id = sd_id

	def format(self, record):
		record.struct_data=self.render_struct_data(record.__dict__.get("_struct_data"))
		return super().format(record)


	def formatTime(self, record, datefmt=None):
		ct = datetime.datetime.fromtimestamp(record.created)
		if datefmt:
			s = ct.strftime(datefmt)
		else:
			t = ct.strftime("%Y-%m-%d %H:%M:%S")
			s = "%s.%03d" % (t, record.msecs)
		return s

	def render_struct_data(self, struct_data):
		if struct_data is None:
			return ""
		else:
			return "[{sd_id} {sd_params}]".format(
				sd_id=self.sd_id,
				sd_params=" ".join(['{}="{}"'.format(key, val) for key, val in struct_data.items()]))



class RFC5424Formatter(StructuredDataFormatter):
	""" This formatter is meant for a SysLogHandler """

	def __init__(self, fmt=None, datefmt=None, style='%'):
		super().__init__(fmt, datefmt, style)
		app_name=Config["logging:formatter_rfc5424"]["app_name"]

		# RFC5424 format
		fmt = '{header} {structured_data} {message}'.format(
			header='{version} {timestamp} {hostname} {app_name} {proc_id} {msg_id}'.format(
				version="1",
				timestamp='%(asctime)s.%(msecs)dZ',
				hostname=socket.gethostname(),
				app_name=app_name,
				proc_id=os.getpid(),
				msg_id='-'),
			structured_data='%(struct_data)s',
			message='%(message)s'
		)

		# Initialize formatter
		super().__init__(
			fmt=fmt,
			datefmt='%Y-%m-%dT%H:%M:%S',
			style='%')

		# Convert time to GMT
		self.converter = time.gmtime



def _loop_exception_handler(loop, context):
	exception = context.pop('exception', None)
	
	message = context.pop('message', '')
	if len(message) > 0:
		message += '\n'

	if len(context) > 0:
		message += pprint.pformat(context)

	if exception is not None:
		ex_traceback = exception.__traceback__
		tb_lines = [ line.rstrip('\n') for line in traceback.format_exception(exception.__class__, exception, ex_traceback)]
		message += '\n' + '\n'.join(tb_lines)

	logging.getLogger().error(message)



if __name__ == '__main__':
	handler = logging.handlers.SysLogHandler()
	handler.setFormatter(RFC5424Formatter())
	handler.setLevel(logging.INFO)
	L = logging.getLogger(__name__)
	L.setLevel(logging.INFO)
	L.addHandler(handler)

	L.info("Info message with structured data", struct_data={"sd1": "test"})
	L.warn("Warn message, with structured data", struct_data={"sd1": "test", "sd2": "test2"})
	L.error("Error message without structured data")
